#include "../snake.h"

/* WEIGHTS */
#define MAXSCORE		100

#define SNAKE_DISTANCE_WEIGHT	.200
#define FRUIT_DISTANCE_WEIGHT	.150
#define RANDOM_WEIGHT		.050
#define INERTIA_WEIGHT		.056
#define SNAKE_PRESENCE_WEIGHT	.300
#define FREE_SPACE_WEIGHT	1

/* dump scores to stderr */
#undef DEBUG

/* prototypes */
chtype look_ahead(struct t_cell *cell, int dir);
static void play (snake *s);

static struct t_cell fruit;

struct t_scored_dir {
	int dir;
	float score;
};


int player_init ()
{
	snake *mysnake = snake_make("ai1", '%', AI);
	mysnake->play = &play;
	register_player(mysnake);
	return 0;
}

static void play(snake *s)
{
	char buf[20];

	/* find fruit position or give up! */
	if ( ! find_fruit() ) {
		fprintf(stderr, "AI can't find the fruit anywhere!\n");
		return;
	}

	/* find best direction */
	find_direction(s);

}


/* return cell in given direction */
chtype look_ahead(struct t_cell *cell, int dir)
{
	int x, y, ch;

	struct t_cell newcell;
	newcell.x = cell->x;
	newcell.y = cell->y;

	/* move new cell */
	move_cell(&newcell, dir, 1);

	ch=mvinch(newcell.y,newcell.x);

	//sprintf(msgbuf, "y: %d, x: %d, ch: %c", y, x, ch&A_CHARTEXT);
	//msg();
	/* return chtype found at new position */
	return ch;
}

				
/* find fruit on board */
int find_fruit()
{

	/* the fruit did not move from previous call */
	if ( IS_FRUIT(mvinch(fruit.y, fruit.x)) ) return 1;

	/* ok, I'll seek it */
	for (fruit.x=1; fruit.x<COLS-1; fruit.x++)
		for (fruit.y=1; fruit.y<LINES-1; fruit.y++)
			if (IS_FRUIT(mvinch(fruit.y,fruit.x))) return 1;

	return 0;
}

/* find the best scored direction */
//int find_direction (struct t_cell *from, struct t_cell *to, int *dir)
int find_direction ( snake *s )
{
	struct t_cell *from;
	int tmpdist, i;
	struct t_scored_dir *scored_dir[4];
	char buf[50];

	from = &(s->head->cell);

	/* init scored_dir array */
	for ( i=0; i<4; i++ )
	{
		scored_dir[i] = malloc(sizeof(struct t_scored_dir));
		scored_dir[i]->dir = i;
		scored_dir[i]->score = 0;
	}

	/* score by snake distance */
	score_by_snake_distance(from, scored_dir);

	/* score by space occupation */
	score_by_free_space (from, scored_dir);

	/* score by fruit distance */
	score_by_fruit_distance(from, &fruit, scored_dir);

	/* score by snake presence */
	score_by_snake_presence (from, &fruit, scored_dir);

	/* score by random ... */
	score_by_random(scored_dir);

	/* score by inertia */
	score_by_inertia(&(s->dir), scored_dir);

	/* sort directions by score */
	sort_direction_by_score (scored_dir);

	/* dump_scores */
#ifdef DEBUG
	dump_scores(s, scored_dir);
#endif

	/* find first free direction (in order of score) */
	for ( i=0; i<4; i++ )
	{
		if ( ! IS_SNAKE(look_ahead(from, scored_dir[i]->dir)) )
		{
			s->dir = scored_dir[i]->dir;
			break;
		}
	}

	/* if no free directions are found, i will be == 4
	 * which is the special 'STOP' direction
	 */

}

int distance ( struct t_cell *from, struct t_cell *to, int dir )
{
	int dist;

	switch (dir)
	{
		case UP:
			dist = from->y - to->y;
			if ( dist < 0 ) dist = ARENA_HEIGHT - abs(dist);
			break;
		case DOWN:
			dist = to->y - from->y;
			if ( dist < 0 ) dist = ARENA_HEIGHT - abs(dist);
			break;
		case LEFT:
			dist = from->x - to->x;
			if ( dist < 0 ) dist = ARENA_WIDTH - abs(dist);
			break;
		case RIGHT:
			dist = to->x - from->x;
			if ( dist < 0 ) dist = ARENA_WIDTH - abs(dist);
			break;
	}

	return dist;
}
			
/* random scoring */
int score_by_random ( dir )
struct t_scored_dir **dir;
{
	int i;
	float weight = RANDOM_WEIGHT;

	i = rand_between(0,3);
	
	dir[i]->score += MAXSCORE * weight;
	//for (i=0; i<4; i++) dir[i]->score += rand_between(0,MAXSCORE) * weight;
}

/* score by inertia */
int score_by_inertia ( int *dir, struct t_scored_dir **scored_dir )
{
	float weight = INERTIA_WEIGHT;

	scored_dir[*dir]->score += MAXSCORE*weight;
}

/* score by snake distance */
int score_by_snake_distance ( from, dir )
struct t_cell *from; struct t_scored_dir **dir;
{
	int i, dist, score;
	float maxdist;
	float weight = SNAKE_DISTANCE_WEIGHT;

	/* horizontal distances */
	maxdist = ARENA_WIDTH;

	/* right */
	dist=0;
	i = from->x;
	do {
		i = ++i > COLS-2 ? 1 : i;
		dist++;
		if ( IS_SNAKE(mvinch(from->y, i)) ) break;
	} while ( i != from->x);
	dir[RIGHT]->score += (dist/maxdist)*MAXSCORE*weight;

	/* left */
	dist=0;
	i = from->x;
	do {
		i = --i < 1 ? COLS-2 : i;
		dist++;
		if ( IS_SNAKE(mvinch(from->y, i)) ) break;
	} while ( i != from->x );
	dir[LEFT]->score += (dist/maxdist)*MAXSCORE*weight;


	/* vertical distances */
	maxdist = ARENA_HEIGHT;

	/* up */
	dist=0;
	i = from->y;
	do {
		i = --i < 1 ? (LINES-2) : i;
		dist++;
		if ( IS_SNAKE(mvinch(i, from->x)) ) break;
	} while ( i != from->y );
	dir[UP]->score += (dist/maxdist)*MAXSCORE*weight;

	/* down */
	dist=0;
	i = from->y;
	do {
		i = ++i > LINES-2 ? 1 : i;
		dist++;
		if ( IS_SNAKE(mvinch(i, from->x)) ) break;
	} while ( i != from->y );
	dir[DOWN]->score += (dist/maxdist)*MAXSCORE*weight;
}

/* score by free space */
int score_by_free_space ( from, dir )
struct t_cell *from; struct t_scored_dir **dir;
{
	int square_side = 3; // be odd!
	struct t_cell cell;
	int half_side;
	int c,d;
	int freespace;
	int maxspace;
	int weight = FREE_SPACE_WEIGHT;

	half_side = (square_side/2);
	maxspace = square_side*square_side;

	/* right */
	freespace = maxspace;
	cell.x = from->x;
	for (
		move_cell(&cell, RIGHT, 1), c=0;
		c<square_side;
		c++, move_cell(&cell, RIGHT, 1 )
	)
	{
		cell.y = from->y;
		for (
			move_cell(&cell, UP, 1), d=0;
			d<square_side;
			d++, move_cell(&cell, UP, 1)
		)
		{
			if ( IS_SNAKE(mvinch(cell.y,cell.x)) ) {
				freespace--;
			}
		}
	}
	dir[RIGHT]->score += freespace * weight;

	/* left */
	freespace = maxspace;
	cell.y = from->y;
	for (
		move_cell(&cell, UP, half_side), c=0;
		c<square_side;
		c++, move_cell(&cell, DOWN, 1)
	)
	{
		cell.x = from->x;
		for (
			move_cell(&cell, LEFT, square_side), d=0; 
			d<square_side;
			d++, move_cell(&cell, RIGHT, 1 )
		)
		{
			if ( IS_SNAKE(mvinch(cell.y,cell.x)) ) freespace--;
		}
	}
	dir[LEFT]->score += freespace * weight;

	/* up */
	freespace = maxspace;
	cell.x = from->x;
	for (
		move_cell(&cell, LEFT, half_side), c=0;
		c<square_side;
		c++, move_cell(&cell, RIGHT, 1)
	)
	{
		cell.y = from->y;
		for (
			move_cell(&cell, UP, 1), d=0;
			d<square_side;
			d++, move_cell(&cell, UP, 1)
		)
		{
			if ( IS_SNAKE(mvinch(cell.y,cell.x)) ) freespace--;
		}
	}
	dir[UP]->score += freespace * weight;

	/* down */
	freespace = maxspace;
	cell.y = from->y;
	for (
		move_cell(&cell, DOWN, 1), c=0;
		c<square_side;
		c++, move_cell(&cell, DOWN, 1)
	)
	{
		cell.x = from->x;
		for (
			move_cell(&cell, LEFT, half_side), d=0;
			d<square_side;
			d++, move_cell(&cell, RIGHT, 1)
		)
		{
			if ( IS_SNAKE(mvinch(cell.y,cell.x)) ) freespace--;
		}
	}
	dir[DOWN]->score += freespace * weight;
}

/* score by snake presence beetween from and to */
int score_by_snake_presence ( from, to, dir )
struct t_cell *from; struct t_cell *to; struct t_scored_dir **dir;
{
	int i, snakefound;
	float weight = SNAKE_PRESENCE_WEIGHT;

	/* right */
	snakefound=0;
	for ( i = from->x+1; i != to->x; i = ++i >= COLS-1 ? 1 : i )
	{
		if ( IS_SNAKE(mvinch(from->y, i)) ) {
			snakefound=1;
			break;
		}
	}
	if ( ! snakefound ) dir[RIGHT]->score += MAXSCORE*weight;

	/* left */
	snakefound=0;
	for ( i = from->x-1; i != to->x; i = --i < 1 ? COLS-1 : i )
	{
		if ( IS_SNAKE(mvinch(from->y, i)) ) {
			snakefound=1;
			break;
		}
	}
	if ( ! snakefound ) dir[LEFT]->score += MAXSCORE*weight;

	/* up */
	snakefound=0;
	for ( i = from->y-1; i != to->y; i = --i < 1 ? LINES-1 : i )
	{
		if ( IS_SNAKE(mvinch(i, from->x)) ) {
			snakefound=1;
			break;
		}
	}
	if ( ! snakefound ) dir[UP]->score += MAXSCORE*weight;

	/* down */
	snakefound=0;
	for ( i = from->y+1; i != to->y; i = ++i >= LINES-1 ? 1 : i )
	{
		if ( IS_SNAKE(mvinch(i, from->x)) ) {
			snakefound=1;
			break;
		}
	}
	if ( ! snakefound ) dir[DOWN]->score += MAXSCORE*weight;
	
}

/* score by fruit distace */
int score_by_fruit_distance(from, to, dir)
struct t_cell *from; struct t_cell *to; struct t_scored_dir **dir;
{
	float maxdist, dist;
	float weight = FRUIT_DISTANCE_WEIGHT;

	/* vertical distances */
	maxdist = ARENA_HEIGHT;

	dist = distance(from, to, DOWN);
	if ( dist ) {
		dir[DOWN]->score +=
		 ( MAXSCORE - ( (dist/maxdist) * MAXSCORE ) ) * weight;
	}

	dist = distance(from, to, UP);
	if ( dist ) {
		dir[UP]->score +=
		( MAXSCORE - ( (dist/maxdist) * MAXSCORE ) ) * weight;
	}

	/* horizontal distances */
	maxdist = ARENA_WIDTH;

	dist = distance(from, to, RIGHT);
	if ( dist ) {
		dir[RIGHT]->score += 
			( MAXSCORE - ( (dist/maxdist) * MAXSCORE ) ) * weight;
	} 

	dist = distance(from, to, LEFT);
	if ( dist ) {
		dir[LEFT]->score += 
		( MAXSCORE - ( (dist/maxdist) * MAXSCORE ) ) * weight;
	}
}


dump_scores ( snake *s, struct t_scored_dir **dirs )
{
	static char Direction[4] = { 'l', 'u', 'r', 'd' };
	char buf[50], buf2[15];
	int i=0;

	sprintf(buf, "%c): ", s->look&A_CHARTEXT);
	for ( i=0; i<4; i++ )
	{
		snprintf(buf2, 15, "%c: %4.2f ",
			Direction[dirs[i]->dir], dirs[i]->score);
		strncat(buf, buf2, 15);
	}

	fprintf(stderr, "%s\n", buf);
	fflush(stderr);
	//msg(buf);
	//refresh();
}

int sort_direction_by_score (struct t_scored_dir **sdir)
{
	int changes, i;
	struct t_scored_dir *ptr;

	do
	{
		changes=0;
		for ( i=0; i<3; i++ )
		{
			if ( sdir[i]->score < sdir[i+1]->score )
			{
				changes++;
				ptr = sdir[i];
				sdir[i] = sdir[i+1];
				sdir[i+1] = ptr;
			}
		}
	} while (changes);
}

msg(char *str)
{
	mvaddstr(0, 20, str);
}


