#include "include/snake.h"
#include <stdlib.h>		/* getenv, ... */
#include <time.h>		/* time (for srand) */
#include <limits.h>		/* for PATH_MAX */
#include <glob.h> 		/* glob_t etc. */
#include <strings.h>		/* strcat, strcmp, ... */
#include <dlfcn.h>		/* dlopen, dlclose, ... */

/* extern data */
extern WINDOW *arenawin;

/* extern functions */
extern int init_graphics (void);
extern void draw_snakes (snake **);
extern void move_snakes (void);
extern void init_fruit (void);
extern int parse_input (void);
extern void draw_motion (void);

/* private functions */
static void snake_play (void);
static int init_snakes (void);
static int load_module (char *);

/* public function */
int register_player(snake *s);
void finish (void);

/* public data */
snake *zoo[MAXPLAYERS];	/* snakes pointer */
int typein;             /* character typed in */
int delay = 40;		/* game delay in milliseconds */
char *rootdir;		/* installation root directory */
unsigned long int turns = 0;	/* game turns */


int main(int argc, char **argv)
{

	/* get installation root directory */
	rootdir = getenv("SNAKEROOT");
	if ( ! rootdir ) {
		fprintf(stderr, "please export SNAKEROOT\n");
		exit(1);
	} 

	/* initialize the graphic or exit */
	if ( ! init_graphics() ) {
		fprintf(stderr, "can't initialize graphic");
		finish();
	}

	/* initialize the random seed */
	srand(time(NULL));

	/* initialize the snakes */
	init_snakes();

	/* initialize the fruit */
	init_fruit();
	
	/* draw the snakes */
	draw_snakes(zoo);

	/* refresh the screen */
	wrefresh(arenawin);

	for (;;)
	{
		/* parse keyboard input */
		if ( !parse_input() ) break;

		/* give players a turn to play */
		snake_play();

		/* move the snakes */
		move_snakes();

		/* draw the snakes */
		draw_snakes(zoo);

		/* show activity */
		draw_motion();

		/* refresh the screen with the new data */
		wrefresh(arenawin);
		
		/* sleep for delay milliseconds */
		napms(delay);

		/* increment bogus time counter */
		turns++;

	}

	finish();
	exit(0); /* finish will exit itself */
}

void finish ()
{
	int i;

	/* reset the screen */
	endwin();

	/* print scores */
	for ( i=0; zoo[i]; i++ ) {
		printf("snake %i (%c): score(%li), length(%i)\n",
			i, (char)zoo[i]->look&A_CHARTEXT,
			zoo[i]->score, zoo[i]->length);
	}

	exit(0);
}

static int init_snakes()
{
	char pattern[PATH_MAX];
	glob_t pglob;
	int i;

	/* create pattern for *.so files */
	strcpy (pattern, rootdir);
	strcat (pattern, "/modules/*.so");

	/* get all *.so files */
	glob (pattern, GLOB_NOSORT, NULL, &pglob);

	for ( i=0; i<pglob.gl_pathc; i++ )
	{
		load_module( pglob.gl_pathv[i] );
	}

	return 1;
}


static int load_module (char *modulename)
{
	const char *dlerr;
	void *module;
	int (*moduleinit)(void);

	fprintf(stderr, "Loading module %s\n", modulename);
	module = dlopen(modulename, RTLD_NOW);
	if (!module) {
		fprintf(stderr, "dlopen(%s): %s\n", modulename, dlerror());
		return 0;
	}

	moduleinit = dlsym(module, "module_init");
        if ( (dlerr = dlerror()) )  {
            fprintf(stderr, "dlsym(module_init): %s\n", dlerr);
	    return 0;
        }

	if ( (*moduleinit)() ) {
		fprintf(stderr, "module initialization failed\n");
		return 0;
	}

	return 1;
}

static void snake_play (void)
{
	int i;

	for (i=0; zoo[i]; i++)
	{
		if ( zoo[i]->died == 2 ) continue;
		(*zoo[i]->play)(zoo[i]);
	}
}


/* player modules must call this function at startup */
int register_player(snake *s)
{
	static int playernum=0;

	/* don't register more then MAXPLAYERS snakes ! */
	/* gotta remove this limitation            */
	if ( playernum >= MAXPLAYERS-1 ) {
		fprintf(stderr, "maximum count of players reached\n");
		return 0;
	}

	/* register new player */
	zoo[playernum++] = s;

	fprintf(stderr,  "player %d registered\n", playernum);
	return playernum-1;
}
